from . import default as default
from . import iter as iter
from . import option as option
from . import ref as ref
from .cfg import cfg as cfg
from .cfg import cfg_attr as cfg_attr
from .default import Default as Default
from .drop import Drop as Drop
from .drop import drop as drop
from .iter import Iter as Iter
from .iter import into_iter as into_iter
from .option import Option as Option
from .option import Some as Some
from .ref import Ref as Ref
from .ref import RefMut as RefMut

__all__: tuple[str, ...]
__url__: str
__author__: str
__version__: str
__docs__: str
__license__: str
